// ignore_for_file: non_constant_identifier_names

library simprokmachine;

import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart';
import 'package:rxdart/rxdart.dart';

// internal types

class _MachineEmit<Input, Output> {
  final Output? output;
  final Handler<Input> setter;

  _MachineEmit.initial(this.setter) : output = null;

  _MachineEmit.normal(this.output, this.setter);
}

class _InternalExpand<Input, Output> {
  final Input? _initial;
  final Machine<Input, Output> _machine;

  _InternalExpand(this._initial, this._machine);
}

class _SubscriptionInput<Input> {
  final Input? _input;

  _SubscriptionInput.initial() : _input = null;

  _SubscriptionInput.normal(Input input) : _input = input;
}

class _InternalFlatMap<MainInput, MainOutput, SubInput, SubOutput> {
  final MainInput? mainInput;
  final MainOutput? mainOutput;
  final SubInput? subInput;
  final SubOutput? subOutput;

  _InternalFlatMap.mainInput(this.mainInput)
      : mainOutput = null,
        subInput = null,
        subOutput = null;

  _InternalFlatMap.mainOutput(this.mainOutput)
      : mainInput = null,
        subInput = null,
        subOutput = null;

  _InternalFlatMap.subInput(this.subInput)
      : mainInput = null,
        mainOutput = null,
        subOutput = null;

  _InternalFlatMap.subOutput(this.subOutput)
      : mainInput = null,
        mainOutput = null,
        subInput = null;
}

class _MergePair<Input, Output> {
  final Set<_MachineEmit<Input, Output>> set;
  final _MachineEmit<Input, Output>? emit;

  _MergePair(this.set, this.emit);
}

class _InternalRenderItem<Input, Output> {
  final RenderItem<Input, Output>? _item;

  _InternalRenderItem.initial() : _item = null;

  _InternalRenderItem.normal(RenderItem<Input, Output> item) : _item = item;
}

class _EmittingInput<Input> {
  final Input? _input;

  _EmittingInput.normal(Input input) : _input = input;

  _EmittingInput.nan() : _input = null;
}

class _EmittingOutput<Input, Output> {
  final Input? _input;
  final Output? _output;

  _EmittingOutput.input(Input input)
      : _input = input,
        _output = null;

  _EmittingOutput.output(Output output)
      : _input = null,
        _output = output;
}

// machines
class _InwardMachine<Input, Output> extends Machine<Input, Output> {
  final Stream<_MachineEmit<Input, Output>> stream;

  _InwardMachine(this.stream);

  @override
  Stream<_MachineEmit<Input, Output>> _stream() {
    return stream;
  }

  static _InwardMachine<ParentInput, ChildOutput>
      create<ParentInput, ChildInput, ChildOutput>(
    Machine<ChildInput, ChildOutput> child,
    Mapper<ParentInput, Inward<ChildInput>> mapper,
  ) {
    Handler<ParentInput> getSetter(
        _MachineEmit<ChildInput, ChildOutput> event) {
      final Handler<ParentInput> result = ((ParentInput input) {
        final mappedInput = mapper(input);

        if (mappedInput._input != null) {
          event.setter(mappedInput._input!);
        } else {
          // do nothing
        }
      });

      return result;
    }

    final Stream<_MachineEmit<ParentInput, ChildOutput>> stream =
        child._stream().map((_MachineEmit<ChildInput, ChildOutput> event) {
      if (event.output == null) {
        // initial
        final Handler<ParentInput> setter = getSetter(event);
        return _MachineEmit<ParentInput, ChildOutput>.initial(setter);
      } else {
        // normal
        final mappedSetter = getSetter(event);

        return _MachineEmit<ParentInput, ChildOutput>.normal(
            event.output!, mappedSetter);
      }
    });

    return _InwardMachine<ParentInput, ChildOutput>(stream);
  }
}

class _OutwardMachine<Input, Output> extends Machine<Input, Output> {
  final Stream<_MachineEmit<Input, Output>> stream;

  _OutwardMachine(this.stream);

  @override
  Stream<_MachineEmit<Input, Output>> _stream() {
    return stream;
  }

  static _OutwardMachine<ChildInput, ParentOutput>
      create<ParentOutput, ChildInput, ChildOutput>(
    Machine<ChildInput, ChildOutput> child,
    Mapper<ChildOutput, Outward<ChildInput, ParentOutput>> mapper,
  ) {
    final Stream<_MachineEmit<ChildInput, ParentOutput>> stream = child
        ._stream()
        .map((_MachineEmit<ChildInput, ChildOutput> event) {
          if (event.output == null) {
            // initial
            return _MachineEmit<ChildInput, ParentOutput>.initial(event.setter);
          } else {
            // normal
            final mappedOutput = mapper(event.output!);
            if (mappedOutput._input != null) {
              // in
              event.setter(mappedOutput._input!);
              return null;
            } else if (mappedOutput._output != null) {
              // out
              return _MachineEmit<ChildInput, ParentOutput>.normal(
                  mappedOutput._output!, event.setter);
            } else {
              // nan
              return null;
            }
          }
        })
        .where((event) => event != null)
        .map((event) => event!);

    return _OutwardMachine<ChildInput, ParentOutput>(stream);
  }
}

class _MergeMachine<Input, Output> extends Machine<Input, Output> {
  final Set<Machine<Input, Output>> _machines;

  _MergeMachine(Set<Machine<Input, Output>> machines)
      : _machines = Set<Machine<Input, Output>>.from(machines);

  @override
  Stream<_MachineEmit<Input, Output>> _stream() {
    final streams = _machines.map((machine) => machine._stream());
    final combined = CombineLatestStream.list(streams);
    final Stream<_MergePair<Input, Output>> scanned =
        combined.scan((accumulated, value, index) {
      if (accumulated.emit == null) {
        // initial
        return _MergePair(
          Set<_MachineEmit<Input, Output>>.from(value),
          _MachineEmit.initial((parentInput) {
            for (var element in value) {
              element.setter(parentInput);
            }
          }),
        );
      } else {
        final newList = value.where((emit1) {
          final filtered =
              accumulated.set.where((emit2) => identical(emit1, emit2));
          final found = (filtered.isEmpty) ? null : filtered.first;
          return found == null;
        });

        if (newList.isEmpty) {
          // something wrong but let's not crash the app
          return accumulated;
        } else {
          final current = newList.first;
          final emit =
              _MachineEmit<Input, Output>.normal(current.output, (parentInput) {
            for (var element in value) {
              element.setter(parentInput);
            }
          });
          return _MergePair(Set<_MachineEmit<Input, Output>>.from(value), emit);
        }
      }
    }, _MergePair({}, null));

    return scanned
        .where((event) => event.emit != null)
        .map((event) => event.emit!);
  }
}

class _EmittingMachine<Input, Output> extends ChildMachine<
    _EmittingInput<Input>, _EmittingOutput<Input, Output>> {
  Handler<_EmittingOutput<Input, Output>>? _callback;

  void send(Input input) {
    if (_callback != null) {
      _callback!(_EmittingOutput<Input, Output>.input(input));
    }
  }

  @override
  void render(_EmittingInput<Input> input,
      Handler<_EmittingOutput<Input, Output>> callback) {
    _callback = callback;
  }
}

class _FlatMapMachine<MainInput, MainOutput, SubInput, SubOutput>
    extends ChildMachine<
        _InternalFlatMap<MainInput, MainOutput, SubInput, SubOutput>,
        _InternalFlatMap<MainInput, MainOutput, SubInput, SubOutput>> {
  Set<Expand<SubInput, SubOutput>> _state = {};
  _EmittingMachine<SubInput, SubOutput>? _emittingMachine;

  StreamSubscription<
      _MachineEmit<_SubscriptionInput<_EmittingInput<SubInput>>,
          _EmittingOutput<SubInput, SubOutput>>>? sub;

  final BiMapper<Set<Expand<SubInput, SubOutput>>, SubInput, Inward<MainInput>>
      _inward;
  final BiMapper<Set<Expand<SubInput, SubOutput>>, MainOutput,
      Set<Expand<SubInput, SubOutput>>> _reduce;

  _FlatMapMachine(this._inward, this._reduce);

  @override
  void render(
      _InternalFlatMap<MainInput, MainOutput, SubInput, SubOutput> input,
      Handler<_InternalFlatMap<MainInput, MainOutput, SubInput, SubOutput>>
          callback) {
    if (input.mainInput != null || input.subOutput != null) {
      // do nothing
    }

    if (input.mainOutput != null) {
      sub?.cancel();
      sub = null;
      _emittingMachine = null;
      final Set<Expand<SubInput, SubOutput>> old =
          Set<Expand<SubInput, SubOutput>>.from(_state);
      _state = Set<Expand<SubInput, SubOutput>>.from(_reduce(
          Set<Expand<SubInput, SubOutput>>.from(old), input.mainOutput!));

      if (_state.isNotEmpty) {
        _emittingMachine = _EmittingMachine<SubInput, SubOutput>();
        final mapped = _state.map((expand) {
          final filtered = old
              .where((element) => identical(expand._machine, element._machine));
          final found = filtered.isEmpty ? null : filtered.first;

          return _InternalExpand<_EmittingInput<SubInput>,
              _EmittingOutput<SubInput, SubOutput>>(
            found == null ? _EmittingInput.normal(expand._initial) : null,
            expand._machine.inward((_EmittingInput<SubInput> input) {
              if (input._input != null) {
                return Inward.Set(input._input!);
              } else {
                return Inward.Nan();
              }
            }).outward((SubOutput output) =>
                Outward.Out(_EmittingOutput.output(output))),
          );
        }).toSet();

        if (_emittingMachine != null) {
          mapped.add(_InternalExpand(_EmittingInput.nan(), _emittingMachine!));
        }

        sub = _start<_EmittingInput<SubInput>,
            _EmittingOutput<SubInput, SubOutput>>(mapped, (output) {
          if (output._input != null) {
            return Inward<_EmittingInput<SubInput>>.Set(
                _EmittingInput<SubInput>.normal(output._input!));
          } else if (output._output != null) {
            callback(_InternalFlatMap<MainInput, MainOutput, SubInput,
                SubOutput>.subOutput(output._output!));
            return Inward<_EmittingInput<SubInput>>.Set(
                _EmittingInput<SubInput>.nan());
          } else {
            return Inward<_EmittingInput<SubInput>>.Nan();
          }
        });
      }
    }

    if (input.subInput != null) {
      final subInput = input.subInput!;
      _emittingMachine?.send(subInput);
      final result =
          _inward(Set<Expand<SubInput, SubOutput>>.from(_state), subInput);
      if (result._input == null) {
        // do nothing
      } else {
        callback(_InternalFlatMap.mainInput(result._input!));
      }
    }
  }
}

class _ExpandMachine<Input, Output> extends ParentMachine<Input, Output> {
  final Machine<Input, Output> _child;

  _ExpandMachine(this._child);

  @override
  Machine<Input, Output> child() {
    return _child;
  }

  static _ExpandMachine<ParentInput, ParentOutput>
      create<ParentInput, ParentOutput, ChildInput, ChildOutput>(
    Machine<ChildInput, ChildOutput> machine,
    BiMapper<Set<Expand<ParentInput, ParentOutput>>, ParentInput,
            Inward<ChildInput>>
        inward,
    BiMapper<Set<Expand<ParentInput, ParentOutput>>, ChildOutput,
            Set<Expand<ParentInput, ParentOutput>>>
        reduce,
  ) {
    final Machine<
        _InternalFlatMap<ChildInput, ChildOutput, ParentInput, ParentOutput>,
        _InternalFlatMap<ChildInput, ChildOutput, ParentInput,
            ParentOutput>> flatted = _FlatMapMachine(inward, reduce);
    final Machine<
        _InternalFlatMap<ChildInput, ChildOutput, ParentInput, ParentOutput>,
        _InternalFlatMap<ChildInput, ChildOutput, ParentInput,
            ParentOutput>> mapped = machine.inward(
        (_InternalFlatMap<ChildInput, ChildOutput, ParentInput, ParentOutput>
            input) {
      if (input.mainInput != null) {
        return Inward.Set(input.mainInput!);
      } else {
        return Inward.Nan();
      }
    }).outward((output) => Outward.Out(_InternalFlatMap.mainOutput(output)));

    final Machine<ParentInput, ParentOutput> merged = merge({flatted, mapped})
        .outward((_InternalFlatMap<ChildInput, ChildOutput, ParentInput,
                ParentOutput>
            output) {
      final Outward<
          _InternalFlatMap<ChildInput, ChildOutput, ParentInput, ParentOutput>,
          ParentOutput> result;
      if (output.mainOutput != null || output.mainInput != null) {
        result = Outward.In(output);
      } else if (output.subOutput != null) {
        result = Outward.Out(output.subOutput!);
      } else {
        // sub input
        result = Outward.Nan();
      }
      return result;
    }).inward((ParentInput input) =>
            Inward.Set(_InternalFlatMap.subInput(input)));

    return _ExpandMachine(merged);
  }
}

// widget machines

class _InwardWidgetMachine<Input, Output> extends WidgetMachine<Input, Output> {
  final Widget _childWidget;
  final Machine<Input, Output> _childMachine;

  _InwardWidgetMachine._(
    Widget childWidget,
    Machine<Input, Output> childMachine,
  )   : _childWidget = childWidget,
        _childMachine = childMachine;

  static _InwardWidgetMachine<ParentInput, ChildOutput>
      create<ParentInput, ChildInput, ChildOutput>(
    WidgetMachine<ChildInput, ChildOutput> child,
    Mapper<ParentInput, Inward<ChildInput>> mapper,
  ) {
    return _InwardWidgetMachine._(
      child._widget(),
      child._machine().inward(mapper),
    );
  }

  @override
  Machine<Input, Output> _machine() {
    return _childMachine;
  }

  @override
  Widget _widget() {
    return _childWidget;
  }
}

class _OutwardWidgetMachine<Input, Output>
    extends WidgetMachine<Input, Output> {
  final Widget _childWidget;
  final Machine<Input, Output> _childMachine;

  _OutwardWidgetMachine._(
    Widget childWidget,
    Machine<Input, Output> childMachine,
  )   : _childWidget = childWidget,
        _childMachine = childMachine;

  static _OutwardWidgetMachine<ChildInput, ParentOutput>
      create<ParentOutput, ChildInput, ChildOutput>(
    WidgetMachine<ChildInput, ChildOutput> child,
    Mapper<ChildOutput, Outward<ChildInput, ParentOutput>> mapper,
  ) {
    return _OutwardWidgetMachine._(
      child._widget(),
      child._machine().outward(mapper),
    );
  }

  @override
  Machine<Input, Output> _machine() {
    return _childMachine;
  }

  @override
  Widget _widget() {
    return _childWidget;
  }
}

class _MergeWidgetMachine<Input, Output> extends WidgetMachine<Input, Output> {
  final WidgetMachine<Input, Output> _main;
  final Set<Machine<Input, Output>> _secondary;

  _MergeWidgetMachine({
    required WidgetMachine<Input, Output> main,
    required Set<Machine<Input, Output>> secondary,
  })  : _main = main,
        _secondary = Set<Machine<Input, Output>>.from(secondary);

  @override
  Machine<Input, Output> _machine() {
    _secondary.add(_main._machine());
    return _MergeMachine(_secondary);
  }

  @override
  Widget _widget() {
    return _main._widget();
  }
}

class _ExpandWidgetMachine<Input, Output> extends WidgetMachine<Input, Output> {
  final Widget _childWidget;
  final Machine<Input, Output> _childMachine;

  _ExpandWidgetMachine._(
    Widget childWidget,
    Machine<Input, Output> childMachine,
  )   : _childWidget = childWidget,
        _childMachine = childMachine;

  static _ExpandWidgetMachine<ParentInput, ParentOutput>
      create<ParentInput, ParentOutput, ChildInput, ChildOutput>(
    WidgetMachine<ChildInput, ChildOutput> child,
    BiMapper<Set<Expand<ParentInput, ParentOutput>>, ParentInput,
            Inward<ChildInput>>
        inward,
    BiMapper<Set<Expand<ParentInput, ParentOutput>>, ChildOutput,
            Set<Expand<ParentInput, ParentOutput>>>
        reduce,
  ) {
    return _ExpandWidgetMachine._(
      child._widget(),
      child._machine().expand(inward: inward, reduce: reduce),
    );
  }

  @override
  Machine<Input, Output> _machine() {
    return _childMachine;
  }

  @override
  Widget _widget() {
    return _childWidget;
  }
}

// start method

StreamSubscription<_MachineEmit<_SubscriptionInput<Input>, Output>>
    _start<Input, Output>(
  Set<_InternalExpand<Input, Output>> expands,
  Mapper<Output, Inward<Input>> mapper,
) {
  final Set<Machine<_SubscriptionInput<Input>, Output>> machines =
      expands.map((_InternalExpand<Input, Output> expand) {
    return expand._machine.inward((_SubscriptionInput<Input> parentInput) {
      if (parentInput._input == null) {
        // initial
        if (expand._initial == null) {
          // no initial input
          return Inward<Input>.Nan();
        } else {
          return Inward<Input>.Set(expand._initial!);
        }
      } else {
        // normal input
        return Inward<Input>.Set(parentInput._input!);
      }
    });
  }).toSet();

  final Machine<_SubscriptionInput<Input>, Output> machine = merge(machines);

  return machine._stream().listen((event) {
    if (event.output == null) {
      // initial
      event.setter(_SubscriptionInput<Input>.initial());
    } else {
      // normal
      final Inward<Input> mapped = mapper(event.output!);
      if (mapped._input == null) {
        // nan. do nothing
      } else {
        event.setter(_SubscriptionInput<Input>.normal(mapped._input!));
      }
    }
  });
}

// Widget

class _RootWidget<Input, Output> extends StatefulWidget {
  final Input _initial;
  final Mapper<Output, Inward<Input>> _map;
  final Widget _widget;
  final Machine<Input, Output> _machine;

  const _RootWidget(
    this._initial,
    this._map,
    this._widget,
    this._machine, {
    Key? key,
  }) : super(key: key);

  @override
  _RootWidgetState<Input, Output> createState() =>
      _RootWidgetState<Input, Output>();
}

class _RootWidgetState<Input, Output>
    extends State<_RootWidget<Input, Output>> {
  StreamSubscription<_MachineEmit<_SubscriptionInput<Input>, Output>>?
      subscription;

  @override
  void initState() {
    subscription = _start<Input, Output>(
      {_InternalExpand<Input, Output>(widget._initial, widget._machine)},
      widget._map,
    );

    super.initState();
  }

  @override
  void dispose() {
    subscription?.cancel();
    subscription = null;
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return widget._widget;
  }
}

// API

// functions
typedef Handler<T> = void Function(T);
typedef BiHandler<T1, T2> = void Function(T1, T2);
typedef TriHandler<T1, T2, T3> = void Function(T1, T2, T3);

typedef Mapper<I, O> = O Function(I);
typedef BiMapper<T1, T2, R> = R Function(T1, T2);
typedef TriMapper<T1, T2, T3, R> = R Function(T1, T2, T3);

// machines

abstract class Machine<Input, Output> {
  Stream<_MachineEmit<Input, Output>> _stream();
}

abstract class ChildMachine<Input, Output> extends Machine<Input, Output> {
  void render(Input input, Handler<Output> callback);

  @override
  Stream<_MachineEmit<Input, Output>> _stream() {
    final PublishSubject<Input?> inputSubject = PublishSubject<Input?>();
    final PublishSubject<Output> outputSubject = PublishSubject<Output>();

    final Stream<_MachineEmit<Input, Output>> inStream =
        inputSubject.startWith(null).asyncExpand((input) {
      if (input != null) {
        return Stream<_MachineEmit<Input, Output>?>.value(null)
            .doOnData(
              (_) =>
                  render(input, ((output) => outputSubject.sink.add(output))),
            )
            .where((event) => event != null)
            .map((event) => event!);
      } else {
        return Stream<_MachineEmit<Input, Output>>.value(
          _MachineEmit<Input, Output>.initial(
              (Input input) => inputSubject.sink.add(input)),
        );
      }
    });

    final Stream<_MachineEmit<Input, Output>> outStream =
        outputSubject.map((output) {
      return _MachineEmit<Input, Output>.normal(
        output,
        (Input input) => inputSubject.sink.add(input),
      );
    });

    return MergeStream([
      inStream,
      outStream,
    ]);
  }
}

abstract class ParentMachine<Input, Output> extends Machine<Input, Output> {
  Machine<Input, Output> child();

  @override
  Stream<_MachineEmit<Input, Output>> _stream() {
    return child()._stream();
  }
}

class BasicMachine<Input, Output> extends ChildMachine<Input, Output> {
  final BiHandler<Input, Handler<Output>> _renderer;

  BasicMachine({
    required BiHandler<Input, Handler<Output>> renderer,
  }) : _renderer = renderer;

  @override
  void render(Input input, Handler<Output> callback) {
    _renderer(input, callback);
  }
}

// widget machines

abstract class WidgetMachine<Input, Output> {
  Machine<Input, Output> _machine();

  Widget _widget();
}

abstract class ChildWidgetMachine<Input, Output>
    extends WidgetMachine<Input, Output> {
  final PublishSubject<RenderItem<Input, Output>> uiSubject = PublishSubject();

  Widget child();

  @override
  Machine<Input, Output> _machine() {
    return BasicMachine(renderer: (input, callback) {
      uiSubject.sink.add(RenderItem(
        input: input,
        callback: (output) => callback(output),
      ));
    });
  }

  @override
  Widget _widget() {
    return StreamProvider<_InternalRenderItem<Input, Output>>(
      create: (_) =>
          uiSubject.map((event) => _InternalRenderItem.normal(event)),
      initialData: _InternalRenderItem.initial(),
      child: child(),
    );
  }
}

abstract class ParentWidgetMachine<Input, Output>
    extends WidgetMachine<Input, Output> {
  late final WidgetMachine<Input, Output> _wm = child();

  WidgetMachine<Input, Output> child();

  @override
  Machine<Input, Output> _machine() {
    return _wm._machine();
  }

  @override
  Widget _widget() {
    return _wm._widget();
  }
}

class BasicWidgetMachine<Input, Output>
    extends ChildWidgetMachine<Input, Output> {
  final Widget _child;

  BasicWidgetMachine({required Widget child}) : _child = child;

  @override
  Widget child() {
    return _child;
  }
}

// extensions

extension InwardMachineExtension<ChildInput, ChildOutput>
    on Machine<ChildInput, ChildOutput> {
  Machine<ParentInput, ChildOutput> inward<ParentInput>(
    Mapper<ParentInput, Inward<ChildInput>> mapper,
  ) {
    return _InwardMachine.create(this, mapper);
  }
}

extension OutwardMachineExtension<ChildInput, ChildOutput>
    on Machine<ChildInput, ChildOutput> {
  Machine<ChildInput, ParentOutput> outward<ParentOutput>(
    Mapper<ChildOutput, Outward<ChildInput, ParentOutput>> mapper,
  ) {
    return _OutwardMachine.create(this, mapper);
  }
}

extension ExpandMachineExtension<ChildInput, ChildOutput>
    on Machine<ChildInput, ChildOutput> {
  Machine<ParentInput, ParentOutput> expand<ParentInput, ParentOutput>({
    required BiMapper<Set<Expand<ParentInput, ParentOutput>>, ParentInput,
            Inward<ChildInput>>
        inward,
    required BiMapper<Set<Expand<ParentInput, ParentOutput>>, ChildOutput,
            Set<Expand<ParentInput, ParentOutput>>>
        reduce,
  }) {
    return _ExpandMachine.create(this, inward, reduce);
  }
}

Machine<Input, Output> merge<Input, Output>(
  Set<Machine<Input, Output>> machines,
) {
  return _MergeMachine(machines);
}

extension InwardWidgetMachineExtension<ChildInput, ChildOutput>
    on WidgetMachine<ChildInput, ChildOutput> {
  WidgetMachine<ParentInput, ChildOutput> inward<ParentInput>(
    Mapper<ParentInput, Inward<ChildInput>> mapper,
  ) {
    return _InwardWidgetMachine.create(this, mapper);
  }
}

extension OutwardWidgetMachineExtension<ChildInput, ChildOutput>
    on WidgetMachine<ChildInput, ChildOutput> {
  WidgetMachine<ChildInput, ParentOutput> outward<ParentOutput>(
    Mapper<ChildOutput, Outward<ChildInput, ParentOutput>> mapper,
  ) {
    return _OutwardWidgetMachine.create(this, mapper);
  }
}

extension ExpandWidgetMachineExtension<ChildInput, ChildOutput>
    on WidgetMachine<ChildInput, ChildOutput> {
  WidgetMachine<ParentInput, ParentOutput> expand<ParentInput, ParentOutput>({
    required BiMapper<Set<Expand<ParentInput, ParentOutput>>, ParentInput,
            Inward<ChildInput>>
        inward,
    required BiMapper<Set<Expand<ParentInput, ParentOutput>>, ChildOutput,
            Set<Expand<ParentInput, ParentOutput>>>
        reduce,
  }) {
    return _ExpandWidgetMachine.create(this, inward, reduce);
  }
}

WidgetMachine<Input, Output> mergeWidgetMachine<Input, Output>({
  required WidgetMachine<Input, Output> main,
  required Set<Machine<Input, Output>> secondary,
}) {
  return _MergeWidgetMachine(main: main, secondary: secondary);
}

// types

class Expand<Input, Output> {
  final Input _initial;
  final Machine<Input, Output> _machine;

  Expand({
    required Input initial,
    required Machine<Input, Output> machine,
  })  : _initial = initial,
        _machine = machine;
}

class Inward<Input> {
  final Input? _input;

  Inward.Set(Input input) : _input = input;

  Inward.Nan() : _input = null;
}

class Outward<ChildInput, ParentOutput> {
  final ChildInput? _input;
  final ParentOutput? _output;

  Outward.In(ChildInput input)
      : _input = input,
        _output = null;

  Outward.Out(ParentOutput output)
      : _input = null,
        _output = output;

  Outward.Nan()
      : _input = null,
        _output = null;
}

class RenderItem<Input, Output> {
  final Input input;
  final Handler<Output> callback;

  RenderItem({
    required this.input,
    required this.callback,
  });
}

// widgets

void runRootMachine<Input, Output>({
  required WidgetMachine<Input, Output> root,
  required Input initial,
  required Mapper<Output, Inward<Input>> map,
}) {
  runApp(
    _RootWidget(
      initial,
      map,
      root._widget(),
      root._machine(),
    ),
  );
}

class MachineConsumer<Input, Output> extends StatelessWidget {
  final Mapper<BuildContext, Widget> _initial;
  final TriMapper<BuildContext, Input, Handler<Output>, Widget> _builder;

  const MachineConsumer({
    Key? key,
    required Mapper<BuildContext, Widget> initial,
    required TriMapper<BuildContext, Input, Handler<Output>, Widget> builder,
  })  : _initial = initial,
        _builder = builder,
        super(key: key);

  @override
  Widget build(BuildContext context) {
    return Consumer<_InternalRenderItem<Input, Output>>(
      builder: (context, value, _) {
        if (value._item == null) {
          // initial
          return _initial(context);
        } else {
          // normal
          return _builder(context, value._item!.input, value._item!.callback);
        }
      },
    );
  }
}
